﻿using EonaCat.HID.EventArguments;
using EonaCat.HID.Models;
using Microsoft.Win32.SafeHandles;
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Runtime.InteropServices;
using System.Threading;
using System.Threading.Tasks;

namespace EonaCat.HID
{
    // This file is part of the EonaCat project(s) which is released under the Apache License.
    // See the LICENSE file or go to https://EonaCat.com/license for full license details.

    internal sealed class HidMac : IHid
    {
        private IntPtr _deviceHandle;

        public HidMac(IntPtr deviceHandle)
        {
            _deviceHandle = deviceHandle;
            CFRetain(_deviceHandle);
        }

        internal void Setup()
        {
            InitializeDeviceInfo();
        }

        public string DevicePath => null;
        public ushort VendorId { get; private set; }
        public ushort ProductId { get; private set; }
        public string SerialNumber { get; private set; }
        public string Manufacturer { get; private set; }
        public string ProductName { get; private set; }

        public int InputReportByteLength { get; private set; } = 64;
        public int OutputReportByteLength { get; private set; } = 64;
        public int FeatureReportByteLength { get; private set; } = 64;

        public bool IsReadingSupport { get; private set; }
        public bool IsWritingSupport { get; private set; }

        public bool IsConnected => _isOpen;

        public IDictionary<string, object> Capabilities { get; private set; } = new Dictionary<string, object>();

        public event EventHandler<HidDataReceivedEventArgs> OnDataReceived;
        public event EventHandler<HidErrorEventArgs> OnError;

        private const int KERN_SUCCESS = 0;
        private const int kIOHIDOptionsTypeNone = 0;

        public void Open()
        {
            if (_isOpen)
            {
                return;
            }

            var result = IOHIDDeviceOpen(_deviceHandle, kIOHIDOptionsTypeNone);
            if (result != KERN_SUCCESS)
            {
                throw new InvalidOperationException($"Failed to open HID device. IOHIDDeviceOpen returned: {result}");
            }

            IsReadingSupport = true;
            IsWritingSupport = CheckOutputReportSupport();

            _isOpen = true;
        }

        private bool CheckOutputReportSupport()
        {
            // On macOS, there's no simple API to test writing directly.
            // You typically assume support based on report length.
            return OutputReportByteLength > 0;
        }

        public void Close()
        {
            IOHIDDeviceClose(_deviceHandle, 0);
            _isOpen = false;
        }

        public void Dispose()
        {
            StopListening();
            Close();
            CFRelease(_deviceHandle);
            _deviceHandle = IntPtr.Zero;
        }

        public void StopListening()
        {
            _listeningCts?.Cancel();
        }

        public async Task WriteOutputReportAsync(HidReport report)
        {
            if (report == null)
            {
                throw new ArgumentNullException(nameof(report));
            }

            if (report.Data == null || report.Data.Length == 0)
            {
                throw new ArgumentException("Data cannot be null or empty", nameof(report));
            }

            await Task.Run(() =>
            {
                // Total length includes reportId + data length
                int length = 1 + report.Data.Length;
                IntPtr buffer = Marshal.AllocHGlobal(length);

                try
                {
                    // First byte is reportId
                    Marshal.WriteByte(buffer, report.ReportId);
                    // Copy the rest of data after the reportId byte
                    Marshal.Copy(report.Data, 0, buffer + 1, report.Data.Length);

                    int res = IOHIDDeviceSetReport(_deviceHandle, 1 /* kIOHIDReportTypeOutput */, report.ReportId, buffer, length);
                    if (res != 0)
                    {
                        OnError?.Invoke(this, new HidErrorEventArgs(this, new Exception($"IOHIDDeviceSetReport (Output) failed: {res}")));
                    }
                }
                finally
                {
                    Marshal.FreeHGlobal(buffer);
                }
            });
        }

        public async Task SendFeatureReportAsync(HidReport report)
        {
            if (report == null)
            {
                throw new ArgumentNullException(nameof(report));
            }

            if (report.Data == null || report.Data.Length == 0)
            {
                throw new ArgumentException("Data cannot be null or empty", nameof(report));
            }

            await Task.Run(() =>
            {
                int length = 1 + report.Data.Length;
                IntPtr buffer = Marshal.AllocHGlobal(length);

                try
                {
                    Marshal.WriteByte(buffer, report.ReportId);
                    Marshal.Copy(report.Data, 0, buffer + 1, report.Data.Length);

                    int res = IOHIDDeviceSetReport(_deviceHandle, 2 /* kIOHIDReportTypeFeature */, report.ReportId, buffer, length);
                    if (res != 0)
                    {
                        OnError?.Invoke(this, new HidErrorEventArgs(this, new Exception($"IOHIDDeviceSetReport (Feature) failed: {res}")));
                    }
                }
                finally
                {
                    Marshal.FreeHGlobal(buffer);
                }
            });
        }

        public async Task<HidReport> GetFeatureReportAsync(byte reportId)
        {
            return await Task.Run(() =>
            {
                int length = FeatureReportByteLength;
                IntPtr buffer = Marshal.AllocHGlobal(length);

                try
                {
                    int res = IOHIDDeviceGetReport(_deviceHandle, 2 /* kIOHIDReportTypeFeature */, reportId, buffer, ref length);
                    if (res != 0)
                    {
                        OnError?.Invoke(this, new HidErrorEventArgs(this, new Exception($"IOHIDDeviceGetReport (Feature) failed: {res}")));
                        return new HidReport(reportId, Array.Empty<byte>());
                    }

                    byte[] outBuf = new byte[length];
                    Marshal.Copy(buffer, outBuf, 0, length);

                    if (outBuf.Length > 0 && outBuf[0] == reportId)
                    {
                        // Data excludes report ID (first byte)
                        var dataOnly = outBuf.Skip(1).ToArray();
                        return new HidReport(reportId, dataOnly);
                    }
                    else
                    {
                        return new HidReport(reportId, outBuf);
                    }
                }
                finally
                {
                    Marshal.FreeHGlobal(buffer);
                }
            });
        }

        public Task<HidReport> ReadInputReportAsync()
        {
            var tcs = new TaskCompletionSource<HidReport>();
            byte[] buffer = new byte[InputReportByteLength];

            InputReportCallback callback = null;

            callback = (ctx, result, sender, report, reportLength) =>
            {
                int length = reportLength.ToInt32();
                if (report == null || length == 0)
                {
                    tcs.TrySetResult(new HidReport(0, Array.Empty<byte>()));
                    return;
                }

                byte[] output = new byte[length];
                Array.Copy(report, output, length);

                byte reportId = output.Length > 0 ? output[0] : (byte)0;

                byte[] reportData;
                if (output.Length > 1)
                {
                    reportData = new byte[output.Length - 1];
                    Array.Copy(output, 1, reportData, 0, reportData.Length);
                }
                else
                {
                    reportData = Array.Empty<byte>();
                }

                var hidReport = new HidReport(reportId, reportData);
                tcs.TrySetResult(hidReport);
            };

            GCHandle.Alloc(callback);
            IOHIDDeviceRegisterInputReportCallback(_deviceHandle, buffer, (IntPtr)buffer.Length, callback, IntPtr.Zero);

            return tcs.Task;
        }

        public Task StartListeningAsync(CancellationToken ct)
        {
            if (_listeningTask != null && !_listeningTask.IsCompleted)
            {
                OnError?.Invoke(this, new HidErrorEventArgs(this, new NotSupportedException("Already listening on this device.")));
                return Task.CompletedTask;
            }

            _listeningCts = CancellationTokenSource.CreateLinkedTokenSource(ct);

            _listeningTask = Task.Run(() =>
            {
                try
                {
                    byte[] buffer = new byte[InputReportByteLength];
                    InputReportCallback callback = (context, result, sender, report, reportLength) =>
                    {
                        int len = reportLength.ToInt32();
                        if (report == null || report.Length < len)
                        {
                            return;
                        }

                        // Extract reportId (first byte)
                        byte reportId = len > 0 ? report[0] : (byte)0;

                        // Extract the rest of the data after reportId
                        byte[] reportData = len > 1 ? report.Skip(1).Take(len - 1).ToArray() : Array.Empty<byte>();

                        var hidReport = new HidReport(reportId, reportData);

                        OnDataReceived?.Invoke(this, new HidDataReceivedEventArgs(this, hidReport));
                    };


                    GCHandle.Alloc(callback);
                    IOHIDDeviceRegisterInputReportCallback(_deviceHandle, buffer, (IntPtr)buffer.Length, callback, IntPtr.Zero);
                    IOHIDDeviceScheduleWithRunLoop(_deviceHandle, CFRunLoopGetCurrent(), IntPtr.Zero);

                    while (!_listeningCts.Token.IsCancellationRequested)
                    {
                        CFRunLoopRun();
                    }

                    CFRunLoopStop(CFRunLoopGetCurrent());
                }
                catch (OperationCanceledException)
                {
                    return;
                }
                catch (NotSupportedException)
                {
                    OnError?.Invoke(this, new HidErrorEventArgs(this, new NotSupportedException("Reading input reports is not supported on this device.")));
                    _listeningCts.Cancel();
                }
                catch (Exception ex)
                {
                    // Handle exceptions during reading
                    if (_listeningCts.IsCancellationRequested)
                    {
                        // Exit if cancellation was requested
                    }
                    OnError?.Invoke(this, new HidErrorEventArgs(this, ex));
                }
            }, _listeningCts.Token);

            return _listeningTask;
        }

        private void InitializeDeviceInfo()
        {
            VendorId = GetIntProperty(kIOHIDVendorIDKey);
            ProductId = GetIntProperty(kIOHIDProductIDKey);
            Manufacturer = GetStringProperty(kIOHIDManufacturerKey);
            ProductName = GetStringProperty(kIOHIDProductKey);
            SerialNumber = GetStringProperty(kIOHIDSerialNumberKey);
        }

        private enum CFStringEncoding : uint
        {
            UTF8 = 0x08000100,
        }

        private ushort GetIntProperty(string key)
        {
            var cfKey = CFStringCreateWithCString(IntPtr.Zero, key, CFStringEncoding.UTF8);
            IntPtr val = IOHIDDeviceGetProperty(_deviceHandle, cfKey);
            CFRelease(cfKey);

            if (val == IntPtr.Zero)
            {
                return 0;
            }

            if (CFGetTypeID(val) == CFNumberGetTypeID())
            {
                int i = 0;
                bool success = CFNumberGetValue(val, CFNumberType.kCFNumberIntType, out i);
                if (success)
                {
                    return (ushort)i;
                }
            }

            return 0;
        }

        private string GetStringProperty(string key)
        {
            var cfKey = CFStringCreateWithCString(IntPtr.Zero, key, CFStringEncoding.UTF8);
            IntPtr val = IOHIDDeviceGetProperty(_deviceHandle, cfKey);
            CFRelease(cfKey);
            if (val == IntPtr.Zero)
            {
                return null;
            }

            string s = CFStringToString(val);
            CFRelease(val);
            return s;
        }

        private string CFStringToString(IntPtr cfStr)
        {
            if (cfStr == IntPtr.Zero)
            {
                return null;
            }

            IntPtr ptr = CFStringGetCStringPtr(cfStr, CFStringEncoding.UTF8);
            if (ptr != IntPtr.Zero)
            {
                // Manually read null-terminated UTF8 string from pointer
                return PtrToStringUTF8Manual(ptr);
            }

            int len = CFStringGetLength(cfStr);
            int max = CFStringGetMaximumSizeForEncoding(len, CFStringEncoding.UTF8) + 1;
            byte[] buf = new byte[max];

            bool success = CFStringGetCString(cfStr, buf, max, CFStringEncoding.UTF8);
            if (success)
            {
                return System.Text.Encoding.UTF8.GetString(buf).TrimEnd('\0');
            }

            return null;
        }

        private string PtrToStringUTF8Manual(IntPtr ptr)
        {
            if (ptr == IntPtr.Zero)
            {
                return null;
            }

            // Find length of null-terminated string
            int len = 0;
            while (Marshal.ReadByte(ptr, len) != 0)
            {
                len++;
            }

            if (len == 0)
            {
                return string.Empty;
            }

            byte[] buffer = new byte[len];
            Marshal.Copy(ptr, buffer, 0, len);
            return System.Text.Encoding.UTF8.GetString(buffer);
        }

        // Constants for IOHIDDevice properties
        private const string kIOHIDVendorIDKey = "VendorID";
        private const string kIOHIDProductIDKey = "ProductID";
        private const string kIOHIDManufacturerKey = "Manufacturer";
        private const string kIOHIDProductKey = "Product";
        private const string kIOHIDSerialNumberKey = "SerialNumber";

        [DllImport("/System/Library/Frameworks/IOKit.framework/IOKit")]
        static extern IntPtr IOHIDDeviceGetProperty(IntPtr dev, IntPtr key);
        [DllImport("/System/Library/Frameworks/IOKit.framework/IOKit")]
        static extern uint IOHIDDeviceOpen(IntPtr dev, uint opt);
        [DllImport("/System/Library/Frameworks/IOKit.framework/IOKit")]
        static extern uint IOHIDDeviceClose(IntPtr dev, uint opt);
        [DllImport("/System/Library/Frameworks/IOKit.framework/IOKit")]
        static extern int IOHIDDeviceSetReport(IntPtr dev, int type, byte id, IntPtr buf, int len);
        [DllImport("/System/Library/Frameworks/IOKit.framework/IOKit")]
        static extern int IOHIDDeviceGetReport(IntPtr dev, int type, byte id, IntPtr buf, ref int len);

        [DllImport("/System/Library/Frameworks/CoreFoundation.framework/CoreFoundation")]
        static extern IntPtr CFStringCreateWithCString(IntPtr alloc, string s, CFStringEncoding enc);
        [DllImport("/System/Library/Frameworks/CoreFoundation.framework/CoreFoundation")]
        static extern void CFRelease(IntPtr cf);
        [DllImport("/System/Library/Frameworks/CoreFoundation.framework/CoreFoundation")]
        static extern IntPtr CFRetain(IntPtr cf);
        [DllImport("/System/Library/Frameworks/CoreFoundation.framework/CoreFoundation")]
        static extern IntPtr CFStringGetCStringPtr(IntPtr str, CFStringEncoding enc);
        [DllImport("/System/Library/Frameworks/CoreFoundation.framework/CoreFoundation")]
        static extern int CFStringGetLength(IntPtr str);
        [DllImport("/System/Library/Frameworks/CoreFoundation.framework/CoreFoundation")]
        static extern int CFStringGetMaximumSizeForEncoding(int len, CFStringEncoding enc);
        [DllImport("/System/Library/Frameworks/CoreFoundation.framework/CoreFoundation")]
        static extern bool CFStringGetCString(IntPtr str, byte[] buf, int bufSize, CFStringEncoding enc);

        [DllImport("/System/Library/Frameworks/CoreFoundation.framework/CoreFoundation")]
        static extern IntPtr CFGetTypeID(IntPtr cf);
        [DllImport("/System/Library/Frameworks/CoreFoundation.framework/CoreFoundation")]
        static extern IntPtr CFNumberGetTypeID();
        [DllImport("/System/Library/Frameworks/CoreFoundation.framework/CoreFoundation")]
        static extern bool CFNumberGetValue(IntPtr num, CFNumberType type, out int value);

        [DllImport("/System/Library/Frameworks/IOKit.framework/IOKit")]
        private static extern void IOHIDDeviceRegisterInputReportCallback(
        IntPtr device,
        byte[] report,
        IntPtr reportLength,
        InputReportCallback callback,
        IntPtr context);

        [DllImport("/System/Library/Frameworks/IOKit.framework/IOKit")]
        private static extern void IOHIDDeviceScheduleWithRunLoop(
        IntPtr device,
        IntPtr runLoop,
        IntPtr runLoopMode);

        [DllImport("/System/Library/Frameworks/CoreFoundation.framework/CoreFoundation")]
        private static extern IntPtr CFRunLoopGetCurrent();

        [DllImport("/System/Library/Frameworks/CoreFoundation.framework/CoreFoundation")]
        private static extern void CFRunLoopRun();

        [DllImport("/System/Library/Frameworks/CoreFoundation.framework/CoreFoundation")]
        private static extern void CFRunLoopStop(IntPtr runLoop);

        // Input report callback delegate
        private delegate void InputReportCallback(
            IntPtr context,
            IntPtr result,
            IntPtr sender,
            byte[] report,
            IntPtr reportLength);

        // For managing lifetime of listening loop
        private CancellationTokenSource _listeningCts;
        private Task _listeningTask;
        private bool _isOpen;

        private enum CFNumberType : int
        {
            kCFNumberIntType = 9
        }
    }
}
